﻿:namespace  Spice
⍝ Spice specific commands Script V1.07

    Cmds←'ULoad' 'UReset' 'UDebug' 'UVersion'  'UClean' 'UMonitor' 'UUpdate' 'URefresh' ⍝ 'UHelp'?

    ⎕ML←1 ⋄ ⎕IO←1 ⋄ NL←⎕av[4 3] ⍝ cannot use ⎕ucs 13 10 on Classic Unix

    rtb←{⍺←' ' ⋄ (-⊥⍨⍵∊⍺)↓⍵} ⋄ DFLAGS←'d1c48vSWwh'

    ∇ r←List
      r←⎕NS¨(⍴Cmds)⍴⊂''
      r.Name←Cmds
      r.(Group Parse)←⊂'Spice' ''
      r[1].Desc←'Load a User command''s code into the workspace'
      r[2].Desc←'Reload all user command definitions in the workspace' ⍝ this cmd is dealt with in SALTUtils
      r[3].(Desc Parse)←'Enable debugging of User Commands'('1s -permanent -flags[∊]',DFLAGS)
      r[4].Desc←'Report various version numbers (e.g. APL/SALT/UCMD)'
      r[5].Desc←'Remove all traces of SALT tags in the ws'
      r[6].(Desc Parse)←'Monitor code execution' '1s -var='
      r[7].Desc←'Update User Commands with the latest version'
      r[8].(Desc Parse)←'Refresh objects in the ws with the most recent version' '1s -noprompt'
    ∇

    ∇ r←Run(Cmd Args);z;la;t;fn;unav;state;ref;cl;setf;oo;un;rel;v;a;noa;fl;nof
      r←0 0⍴''
      :Select Cmd
     
      :Case 'ULoad'
          r←'No such command found' ⋄ v←1↑⍴t←⎕SE.cbbot.bandsalt.Spice.List
          un←⎕SE.Dyalog.Utils.lcase¨t[;1]
          :If ∨/z←un∊⊂cl←⎕SE.Dyalog.Utils.lcase ref←Args~' '
          :OrIf 1=+/z←cl∘≡¨(⍴cl)↑¨un
              r←⍕⎕SE.SALT.Load'"',(6⊃fn←t[z⍳1;]),'" -target=',⍕⎕SE.SALTUtils.DF ##.THIS
              r←'Command "',(1⊃fn),'" is now found in <',r,'>'
          :EndIf
     
      :Case 'UDebug'
          r←'Is ',state←(0 1⍳⎕SE.SALTUtils.DEBUG)⊃'OFF' 'ON' 'ON+'
          →0/⍨(noa←' '∧.=a←⍕Args.Arguments)∧nof←0≡fl←Args.flags
          '-permanent does not apply to flags'⎕SIGNAL 11/⍨Args.permanent>nof
          r←'Unable to set (enter ON or OFF)'
          :If ∨/t←'off' 'on' ''∊⊂⎕SE.Dyalog.Utils.lcase⌽∘rtb⍣2,a~' '
              :If nof
                  r←'Was ',state
                  ⎕SE.SALTUtils.DEBUG←t[2]
                  :If Args.permanent ⋄ {}⎕SE.SALT.Settings'debug ',(⍕t[2]),' -p' ⋄ :EndIf
              :Else
         ⍝ Set the interpreter (DEBUG) flags (OFF means remove them, ON adds them, neither queries)
                  ((fl='d')/fl)←'1'  ⍝ d ≡ 1 flag
                  r←(1+¯1↑t)⊃'Was: ' 'Is: '
                  setf←{2 ⎕NQ'.' 'SetDFlags'⍵}
                  r←r,t[2]{n←2⍴⍨⍴tags←⌽1↓DFLAGS
                      old←bits←n⊤setf 0  ⍝ find current state
                      ((tags∊⍵)/bits)←⍺
                      x←setf 2⊥bits      ⍝ reset to new value
                      1⌽'""',old/tags}fl,(fl≡1)/DFLAGS
              :EndIf
          :EndIf
     
      :Case 'UReset'
          ⍝ Execution of UReset is handled in the framework; appears here for the HELP only
          ⍝ and in case a shorter form of 'ureset' has been entered:
          t←⍕1↑⍴⎕SE.cbbot.bandsalt.Spice.List←2⊃⎕SE.SALTUtils.GetSpiceList'*'
          r←t,' commands reloaded'
     
      :Case 'UVersion'
          :If 0<⍴fn←{(+/∧\' '=⍵)↓⍵}Args ⍝ are we trying to find the version of a ws?
              :Trap 22
                  t←('"'~⍨fn,(~'.'∊fn)/'.DWS')⎕NTIE 0 ⋄ r←0.1⊥⌽⎕NREAD t,83 2 4 ⋄ ⎕NUNTIE t
              :Else
                  'No such workspace'⎕SIGNAL 22
              :EndTrap
          :Else
              r←⊂'APL'('.'⎕WG'aplversion'),'Classic' 'Unicode'[1+80=⎕DR''],⊂{0::'?' ⋄ 2 ⎕NQ'.'⍵}'getbuildid'
              t←unav←'(unavailable)'
              :If 9=⎕SE.⎕NC'SALT' ⋄ t←⎕SE.SALT.Version ⋄ :EndIf
              r,←⊂'SALT't
              r,←⊂'UCMD'⎕SE.SALTUtils.UVersion
              t←unav
              :Trap 6
                  ⎕USING←'System' ⋄ t←{(2>+\'.'=⍵)/⍵}⍕Environment.Version
              :EndTrap
              r,←⊂'.NET't
         ⍝ Other info such as computer name, OS, 64bOS, #CPUs could be added
              r←((⍴r),1)⍴r
          :EndIf
     
      :Case 'UClean'
          ⎕CS #
          ⎕SE.SALTUtils.cleanWS #
          ⎕←'* Cleanup done'
     
      :Case 'UMonitor'
          state←(0 1⍳⎕SE.SALTUtils.MONITOR)⊃'OFF' 'ON' 'ON+'
          :If t←0≡v←Args.var ⋄ r←'Is ',state
          :Else ⋄ rel←~'#⎕'∊⍨1↑v
              'MONITOR cannot be used from a temp NS'⎕SIGNAL 11/⍨rel∧'['∊t←⍕##.THIS
              v←(rel/t,'.'),v
              'invalid variable name'⎕SIGNAL 11↓⍨(⎕NC v)∊0 2
              ⎕SE.SALTUtils.MONITORNAME←v
              r←'''',v,''' will be used for monitoring'
              ⎕SE.SALTUtils.MONITOR←1
          :EndIf
          →0↓⍨⍴Args.Arguments
          :If ∨/oo←'off' 'on'∊⎕SE.Dyalog.Utils.lcase¨Args.Arguments
              r←'Was ',state
              :If ⎕SE.SALTUtils.MONITOR←oo[2] ⍝ turn ON? clear data
              :AndIf 2=⎕NC t←⎕SE.SALTUtils.MONITORNAME
                  ⍎t,'←⍬'
              :EndIf
          :Else ⋄ r←'Unable to set (enter ON or OFF)'
          :EndIf
     
      :Case 'UUpdate'
          r←Update Cmd Args
     
      :Case 'URefresh'
          r←(~Args.noprompt)refreshWS #
     
      :EndSelect
    ∇

    isYes←{~⍺:1 ⋄ t←⍴⍞←⍵ ⋄ ∨/'yY'∊t↓⍞}

    ∇ {n}←confirm refreshWS tgt;tgt;lst;sons;file;srcws;t;b;obj;cload;call1
    ⍝ This fn is used to update all objects starting at tgt
      n←0 ⍝ number of changes made
      :If call1←0=⎕NC'LIST' ⋄ ⎕SHADOW'LIST' ⋄ LIST←⍬ ⋄ :EndIf ⍝ avoid recursion, keep track of nss done
      LIST,←tgt
     
      cload←{(o f p)←⍵ ⋄ 922 22::0⊣⎕←'*** File ',f,' NOT FOUND for object ',o ⋄ ⍺≢⎕SE.SALT.Load f,p}
     
      :If 0≢srcws←{0::0 ⋄ ⎕SRC ⍵}tgt ⍝ chk scripted nss
          :If 9=tgt.⎕NC'SALT_Data'
              :If srcws cload tgt(file←tgt.SALT_Data.SourceFile)' -source'
              :AndIf confirm isYes⍕'** Confirm overwrite ',tgt,' with contents of file ',file,'  '
                  n+←⍴,⎕←0 tgt.## ⎕SE.SALT.Load file ⍝ we need to redo the load to tag the object
              :EndIf
          :EndIf
      :Else ⍝ must be a regular ns
     
          :If 2=#.⎕NC'SALT_Var_Data.VD'
          :AndIf 1∊b←#.SALT_Var_Data.VD[;⎕IO]{⍵∘≡¨{⍵↓⍨-1+⊥⍨'.'≠⍵}¨⍺}{df←⍵.⎕DF ⎕NULL ⋄ (⍵.⎕DF df)⊢⍕⍵}tgt
              :For obj file :In ↓b⌿#.SALT_Var_Data.VD[;⍳2]
                  :If (⍎obj)cload obj file' -noname -nolink'
                  :AndIf confirm isYes⍕'** Confirm overwrite ',obj,' with contents of file ',file,'  '
                      n+←⍴,⎕←0 tgt ⎕SE.SALT.Load file ⍝ we need to redo the load to tag the object
                  :EndIf
              :EndFor
          :EndIf
     
          :For obj :In lst←tgt.⎕NL-3 4
              :If 2≡≡srcws←tgt.⎕NR obj
              :AndIf ~0∊⍴t←⎕SE.SALTUtils.fnData srcws obj
              :AndIf srcws cload obj t.File' -source'
              :AndIf confirm isYes⍕'** Confirm overwrite ',obj,' with contents of file ',file,'  '
                  n+←⍴,⎕←0 tgt ⎕SE.SALT.Load file
              :EndIf
          :EndFor
     
          :If 0<⍴lst←tgt.⎕NL-9
          :AndIf 0<⍴lst←LIST~⍨tgt⍎¨lst
            ⍝ Recurse thru non-scripted children
              :If 0<⍴sons←{⍵/⍨~1∊¨(⍕¨⍵)∊¨⊂'[('}lst
                  n+←+/∊confirm refreshWS¨sons
              :EndIf
          :EndIf ⍝ all nss
      :EndIf
     
      :If call1 ⋄ n←(⍕n),' objects refreshed' ⋄ :EndIf
    ∇


    ∇ r←Update(Cmd Args);t;upd;del;ec;⎕TRAP;msg
    ⍝ Cover for update procedure
      :If 0<⎕SE.SALTUtils.DEBUG
      :AndIf 9=#.⎕NC'update'
          upd←#.update ⍝ use ws version in debug mode
      :Else
          (↑'DRC' 'HTTPUtils')⎕CY'conga'
          (ec msg t)←3↑HTTPGet'http://tools.dyalog.com/UCMDS/update.dyalog'
          ⎕TRAP←((900+⍳12)'N')(0 'C' '→ERR')
          {}÷0∊ec
          del←⎕AV[4 3]∩t←'UTF-8'⎕UCS ⎕UCS 3↓t ⍝ we cannot use ⎕UCS 13 in Classic Unix
          t←{1↓¨⍵⊂⍨⍵∊1↑⍵}(1↑del),t~1↓del
          upd←0 ⎕FIX t
      :EndIf
      :If ' '∊1↑0⍴Args ⍝ character argument?
          r←upd.Run(Cmd Args)
      :Else
          r←Args upd.Help(Cmd)
      :EndIf
      →0
     ERR:⎕TRAP←0⍴⎕TRAP
      r←'Update unavailable: ',(⎕IO+0=ec)⊃(msg)(⎕EM ⎕EN)
    ∇

    split←{(p↑⍵)((p←¯1+⍵⍳⍺)↓⍵)}

    ∇ r←{certs}HTTPGet url;U;protocol;wr;key;flags;pars;secure;data;z;header;datalen;host;port;done;cmd;b;page;auth;p;x509;priority
⍝ Get an HTTP page, format [HTTP[S]://][user:pass@]url[:port][/page]
⍝ Opional Left argument: PublicCert PrivateKey SSLValidation
⍝ Makes secure connection if left arg provided or URL begins with https:
     
⍝ Result: (return code) (HTTP headers) (HTTP body) [PeerCert if secure]
     
      U←HTTPUtils ⍝ Uses utils from here
      {}DRC.Init''
     
      p←(∨/b)×1+(b←'//'⍷url)⍳1
      secure←(2=⎕NC'certs')∨(U.lc(p-2)↑url)≡'https:'
      port←(1+secure)⊃80 443 ⍝ Default HTTP/HTTPS port
      url←p↓url              ⍝ Remove HTTP[s]:// if present
      host page←'/'split url,(~'/'∊url)/'/'    ⍝ Extract host and page from url
     
      :If 0=⎕NC'certs' ⋄ certs←'' ⋄ :EndIf
     
      :If secure
          x509 flags priority←3↑certs,(⍴,certs)↓(⎕NEW DRC.X509Cert)32 'NORMAL:!CTYPE-OPENPGP'  ⍝ 32=Do not validate Certs
          pars←('x509'x509)('SSLValidation'flags)('Priority'priority)
      :Else ⋄ pars←''
      :EndIf
     
      :If '@'∊host ⍝ Handle user:password@host...
          auth←NL,'Authorization: Basic ',(U.Encode(¯1+p←host⍳'@')↑host)
          host←p↓host
      :Else ⋄ auth←''
      :EndIf
     
      host port←port U.HostPort host ⍝ Check for override of port number
     
      :If 0=⊃(r cmd)←2↑DRC.Clt''host port'Text' 100000,pars ⍝ 100,000 is max receive buffer size
      :AndIf 0=⊃r←DRC.Send cmd('GET ',page,' HTTP/1.1',NL,'Host: ',host,':',(⍕port),NL,'Accept: */*',auth,NL,NL)
     
          done data header←0 ⍬(0 ⍬)
     
          :Repeat
              :If ~done←0≠1⊃wr←DRC.Wait cmd 5000            ⍝ Wait up to 5 secs
                  :If wr[3]∊'Block' 'BlockLast'                ⍝ If we got some data
                      :If 0<⍴data,←4⊃wr
                      :AndIf 0=1⊃header
                          header←U.DecodeHeader data
                          :If 0<1⊃header
                              data←(1⊃header)↓data
                              datalen←⊃((2⊃header)U.GetValue'Content-Length' 'Numeric'),¯1 ⍝ ¯1 if no content length not specified
                          :EndIf
                      :EndIf
                  :Else
                      ⎕←wr ⍝ Error?
                      ∘
                  :EndIf
     
                  done←'BlockLast'≡3⊃wr                        ⍝ Done if socket was closed
                  :If datalen>0
                      done←done∨datalen≤⍴data ⍝ ... or if declared amount of data rcvd
                  :Else
                      done←done∨(∨/'</html>'⍷data)∨(∨/'</HTML>'⍷data)
                  :EndIf
              :EndIf
          :Until done
     
          :Trap 0 ⍝ If any errors occur, abandon conversion
              :If ∨/'charset=utf-8'⍷(2⊃header)U.GetValue'content-type' ''
                  data←'UTF-8'⎕UCS ⎕UCS data ⍝ Convert from UTF-8
              :EndIf
          :EndTrap
     
          r←(1⊃wr)(2⊃header)data
          :If secure ⋄ r←r,⊂DRC.GetProp cmd'PeerCert' ⋄ :EndIf
      :Else
          r←r,⊂'Connection failed (',cmd,')'
      :EndIf
     
      z←DRC.Close cmd
      DRC.⎕EX¨DRC.⍙naedfns ⍝ we need to do this otherwise the connection stays opened
    ∇

    ∇ r←level Help Cmd
      :Select Cmd
      :Case 'ULoad'
          r←,⊂'Load the class associated with a User cmd'
          →level↓end
          r,←'' 'Example, load the script where the SAVE cmd resides:' '      ]uload save' 'Command "save" now found in #.SaltInSpice'
      :Case 'UReset'
          r←⊂'Forces a reload of all user commands, to ensure that parsing information'
          r,←⊂'is up-to-date (after changing parsing rules for a command).'
      :Case 'UDebug'
          r←,⊂']UDebug on       Enables debugging of user commands'
          r,←⊂']UDebug off      Causes all errors in user commands to be trapped (this is the default)'
          r,←⊂']UDebug on -flag=X Sets the APL debug flag X ON (debug flags are ',DFLAGS,')'
          r,←⊂']UDebug  -flags  reports the current debug flags settings'
          r,←'' 'NB: With debugging enabled, a trailing "-" on a user command allows easy tracing of the call'
          →level↓end
          r,←'' 'Examples:' 'Add the W and v flags to be turned on:' '      ]udebug on +fl=Wv' 'Was: ""'
          r,←'Remove the "v" flag:' '      ]udeb off -f=v' 'Was: "Wv"' 'Find which flags are ON:' '      ]ud -flags' 'Is: "W"'
      :Case 'UVersion'
          r←⊂'Reports various version numbers (APL, SALT, UCMD, .NET, etc.)'
      :Case 'UClean'
          r←⊂'Removes all the SALT tags used to track source'
      :Case 'UMonitor'
          r←'Turn monitoring ON or OFF' 'Results are set in #.UCMDMonitor after each invocation of a cmd.'
          r,←⊂'-var= sets the name of the variable to store the result'
          :If ×level
              r,←⊂'If var is specified monitoring is turned on automatically'
              r,←⊂'Turning monitoring ON discards any previous data'
          :EndIf
      :Case 'UUpdate'
          r←Update Cmd level
      :Case 'URefresh'
          r←⊂'Refreshes all SALTed objects by reloading the lastest version.'
          r,←⊂'-noprompt skips the confirmation'
      :EndSelect
     end:r←↑r
    ∇

:Endnamespace