﻿:Namespace APLTreeUtils
⍝ Homepage: http://aplwiki.com/APLTreeUtils
⍝
⍝ *** Version 2.3.5 ⋄ 2012-10-11 ***
⍝ 2.3.5  Bug in ""FindPathTo"" fixed.
⍝ 2.3.4  WriteUtf8File crashed in case "append" was specified and the file did not exist.
⍝ 2.3.3  Fix in 2.3.1 was faulty: when the file did not already existed it wasn't created anymore.
⍝
⍝ Needs Dyalog Version 12, Unicode for "WriteUtf8File" & "ReadUtf8File"
⍝ From version 1.32 on, all fns in this script are tested to be _
⍝ independent from ⎕IO and ⎕ML except when declared to be different (Where..,)

    ∇ array←Uppercase array;ToUpper;⎕ML;⎕IO
      ⎕IO←⎕ML←1
      :If ~0∊⍴array
          'ToUpper'⎕NA'I4 USER32.C32|CharUpper',('*A'⊃⍨1+12>{⍎⍵↑⍨¯1+⍵⍳'.'}2⊃'.'⎕WG'APLVersion'),' =0T'
          :If ~0 1∊⍨≡array
              array←(ToUpper Trans)¨array
          :Else
              array←(ToUpper Trans)array
          :EndIf
      :EndIf
    ∇

    ∇ array←Lowercase array;ToLower;⎕ML;⎕IO
      ⎕IO←⎕ML←1
      :If ~0∊⍴array
          'ToLower'⎕NA'I4 USER32.C32|CharLower',('*A'⊃⍨1+12>{⍎⍵↑⍨¯1+⍵⍳'.'}2⊃'.'⎕WG'APLVersion'),' =0T'
          :If ~0 1∊⍨≡array
              array←(ToLower Trans)¨array
          :Else
              array←(ToLower Trans)array
          :EndIf
      :EndIf
    ∇

      Trans←{⎕IO←1
      ⍝ This operator can produce a WS FULL that often disappears when →⎕LC is executed.
      ⍝ Therefore it calles itself via a WS FULL guard but just once.
          ⍺←0               ⍝ The default.
          ⍺:(⍴⍵)⍴2⊃⍺⍺⊂,⍵    ⍝ If ⍺ is true then by now a workspace compaction was performed.
          1::1 ∇ ⍵          ⍝ In case of a WS FULL call itself with a 1 as left argument.
          (⍴⍵)⍴2⊃⍺⍺⊂,⍵      ⍝  Execute it.
      }

    IsChar ←{0 2∊⍨10|⎕dr ⍵}  ⍝ Version 12 compatible: is 82=⎕dr in V11

    ∇ r←IsUnicode
    ⍝ Return a 1 if running under Dyalog Unicode
      r←80=⎕DR' '
    ∇

      SplitPath←{
          ⎕ML←⎕IO←1
          ⍺←'/\'
          l←1+-⌊/⍺⍳⍨⌽⍵
          (l↓⍵)(l↑⍵)
      }

      Split←{
          ⎕ML←⎕IO←1
          ⍺←⎕UCS 13 10 ⍝ Default is CR+LF
          (⍴,⍺)↓¨⍺{⍵⊂⍨⍺⍷⍵}⍺,⍵
      }

      Last←{
          ⎕ML←⎕IO←1
          ⍺←'.'
          (⍴,⍵)=where←¯1+⌊/⍺⍳⍨⌽,⍵:0⍴⍵
          0 1∊⍨≡r←(-where)↑⍵:r
          ''⍴r
      }

      Nest←{
          ⎕ML←⎕IO←1
          (⊂∘,⍣(0 1∊⍨≡⍵))⍵
      }

    Enlist←{⎕ml←3 ⋄ ∊⍵}

      dmb←{
      ⍝ delete leading, trailing and multiple blanks from scalar or vector.
          ⍺←' '
          ⎕ML←⎕IO←1
          ~0 1∊⍨≡⍵:∇¨⍵
          2=⍴⍴⍵:Mix ∇¨↓⍵
          (,⍺)≡,⍵:''
          w←1↓¯1↓⍺{⍵/⍨~(2⍴⍺)⍷⍵}⍺,⍵,⍺
          (0=⍴⍴⍵)∧1=⍴w:⍬⍴⍵
          w
      }

      dlb←{
      ⍝ Delete leading blanks. Accepts scalar, vector and matrix as argument.
          (2=|≡⍵):∇¨⍵
          (1=⍴⍴⍵):(+/∧\' '=⍵)↓⍵                 ⍝ Vectors (main application)
          (2=⍴⍴⍵):(+/∧\' '=⍵)⌽⍵                 ⍝ Matrix
          (0=⍴⍴⍵):(⎕IO+' '≡⍵)⊃⍵''               ⍝ Scalar
          'Invalid argument'⎕SIGNAL 11
      }

      dtb←{
      ⍝ Delete trailing blanks. Accepts scalar, vector and matrix as argument.
          (2=|≡⍵):∇¨⍵
          (1=⍴⍴⍵):⌽{(+/∧\' '=⍵)↓⍵}⌽⍵            ⍝ Vectors (main application)
          (2=⍴⍴⍵):(-+/∧⌿∧\' '=⌽⍵)↓[2]⍵          ⍝ Matrix
          (0=⍴⍴⍵):(⎕IO+' '≡⍵)⊃⍵''               ⍝ Scalar
          'Invalid argument'⎕SIGNAL 11
      }

    ∇ r←{length}FormatDateTime ts;ts2;formatstring;bool;⎕IO;⎕ML;buffer
⍝ Format the right argument (defaults to ⎕TS if empty) as a string with:
⍝ 'YYYY-MM-DD HH:MM:SS.MILLISECOND
⍝ If the right argument has not 7 but 6 or 3 elements, formatting is done accordingly.
⍝ Via the left argument the length of the right argument can be set to _
⍝ ⍬, 3, 6 or 7; default is 6; ⍬ Accepts any length of the right argument which is 3,6 or 7.
⍝ If the right argument is a simple vector, a string is returned.
⍝ If it is a matrix, a matrix is returned.
      ⎕IO←⎕ML←1
      :If ⍬≡length←{2=⎕NC ⍵:⍎⍵ ⋄ 6}'length'
          length←''⍴¯1↑⍴ts
      :EndIf
      :If ~0∊⍴ts
          :If 2=⍴⍴ts
              buffer←{⍵/⍨0<⊃∘⍴¨⍵~¨⊂' ' 0}↓ts
              'Invalid right argument: must be integer'⎕SIGNAL 11/⍨163≠∪⎕DR¨buffer
              'Invalid right argument: must not be negative'⎕SIGNAL 11/⍨∨/¯1∊¨×¨buffer
              'Invalid right argument: must be simple'⎕SIGNAL 11/⍨1≠∪≡¨buffer
          :Else
              'Invalid right argument: must be integer'⎕SIGNAL 11/⍨163≠⎕DR ts
              'Invalid right argument: must not be negative'⎕SIGNAL 11/⍨¯1∊×,ts
              'Invalid right argument: must be simple'⎕SIGNAL 11/⍨1≠≡ts
          :EndIf
      :EndIf
      :If 2=⍴⍴ts
          ts2←length↑[2]ts
      :Else
          ts2←,[0.5]length↑{0∊⍴⍵:⎕TS ⋄ ts}ts
      :EndIf
      :Select First length
      :Case 3
          formatstring←'ZI4,<->,ZI2,<->,ZI2'
      :CaseList 6 7
          formatstring←'ZI4,<->,ZI2,<->,ZI2,< >,ZI2,<:>,ZI2,<:>,ZI2'
      :Else
          'Invalid left argument'⎕SIGNAL 11
      :EndSelect
      bool←(ts2∨.≠' ')∧ts2∨.≠0
      r←bool⍀formatstring ⎕FMT(6⌊length)↑[2]bool⌿ts2
      :If 7=2⊃⍴ts2
          r←⊃(↓r),¨{0=⍵:'' ⋄ 0∊⍴⍵~' ':'' ⋄ '.',⍕⍵}¨ts2[;7]
      :EndIf
      :If 2≠⍴⍴ts
          r←,r
      :EndIf
    ∇

    ∇ R←CreateUUID;∆UuidCreate;⎕ML;⎕IO
      ⎕ML←⎕IO←1
      '∆UuidCreate'⎕NA'I RPCRT4|UuidCreate >{I1[4] I1[2] I1[2] I1[2] I1[6]}'
      R←2⊃∆UuidCreate 1
      R←'0123456789ABCDEF{-'[1+∊16 17 17 17 17,¨∊¨⍉¨16 16∘⊤¨R],'}'
    ∇

    ∇ r←{flat}ReadUtf8File filename;fno;noOfBytes;bytes;⎕IO;⎕ML;b
    ⍝ This function works also with the Classic version if the file to be read does not contain "real" Unicode data.
    ⍝ By defaults ""ReadUtf8File"" returns a nested vector were each item carries a record.
    ⍝ If you want to get a simple stream specify 'flat' as left argument.
      ⎕IO←⎕ML←1
      r←''
      flat←{0=⎕NC ⍵:0 ⋄ 'flat'≡⍎⍵}'flat'
      :Trap 19 22
          fno←(filename~'"')⎕NTIE 0
      :Else
          ('Could not read file: ',filename)⎕SIGNAL ⎕EN
      :EndTrap
      noOfBytes←⎕NSIZE fno
      bytes←⎕NREAD fno 83,noOfBytes
      ⎕NUNTIE fno
      bytes+←256×bytes<0                 ⍝ Make sure it is unsigned
      bytes↓⍨←3×239 187 191≡3⍴bytes      ⍝ drop a potential UTF-8 marker
      r←'UTF-8'⎕UCS bytes
      :If ~flat
          :If ∨/(⎕UCS 13 10)⍷r
              r←Split r
          :ElseIf ∨/r=⎕UCS 10
              r←(⎕UCS 10)Split r
          :EndIf
      :EndIf
    ∇

    ∇ {r}←{append}WriteUtf8File(file data);fno;fullname;flag;⎕ML;⎕IO;i;max;size;simpleFlag;wasOpenFlag
     ⍝ Write UTF-8 "data" to "file" (WITHOUT a BOM!).
     ⍝ If the left argument is string "append" then "data" is appended _
     ⍝ to an already existing file. If there is no such file yet _
     ⍝ it is created no matter what the left argument is.
     ⍝ In case of "append" the "file" parameter, which is normally _
     ⍝ expected to be a filename, can also be a tie number.
     ⍝ When the tie fails the function tries a couple of times with an
     ⍝ increasing delay before giving up.
     ⍝ Note that when a nested vector is passed as right argument then _
     ⍝ CR+LF are added to the end of each item except the very last one. _
     ⍝ When "append" is specified, CR+LF are also added as a prefix to _
     ⍝ the data in case the file does already exist and was not empty yet.
     ⍝ If a simple string is passed it is written as it is: nothing is _
     ⍝ added at all.
     ⍝ "data" may be either
      ⎕IO←⎕ML←1
      r←''
      'Invalid data: must not be a matrix'⎕SIGNAL 11/⍨2=⍴⍴data
      append←{2=⎕NC ⍵:⍎⍵ ⋄ ''}'append'
      'Invalid right argument'⎕SIGNAL 11/⍨~(⊂append)∊'append' 1 0 ''⍬
      append←(⊂append)∊'append' 1
      flag←0
      :If ~simpleFlag←0 1∊⍨≡data
          data←¯2↓⊃,/data,¨⊂⎕UCS 13 10
      :EndIf
      max←5
      i←0
      size←0
      fno←⍬
      :Repeat
          ⎕DL 1×i>0
          :If wasOpenFlag←0=1↑0⍴file
          :AndIf append
              fno←file
              size←⎕NSIZE fno
              flag←1
          :EndIf
          :If ~wasOpenFlag
              :Trap 19 22
                  fno←(file~'"')⎕NTIE 0 17  ⍝ Open exclusively
                  size←⎕NSIZE fno
                  flag←1
              :Case 22
                  flag←1                    ⍝ That's just fine
              :Else
                  ('Could not open file ',file)⎕SIGNAL ⎕EN
              :EndTrap
          :EndIf
      :Until flag∨max<i←i+1
      :If append
          :If (0<size)∧~simpleFlag
              data,⍨←⎕UCS 13 10
          :EndIf
      :EndIf
      :If ~append
      :OrIf append∧fno≡⍬
          :Trap 0 ⋄ (file~'"')⎕NERASE fno ⋄ :EndTrap
          fno←(file~'"')⎕NCREATE 0
      :EndIf
      data←⎕UCS'UTF-8'⎕UCS data             ⍝ Enforce UTF-8
      data ⎕NAPPEND fno
      :If ~wasOpenFlag
          ⎕NUNTIE fno
      :EndIf
    ∇

    ∇ {r}←GoToWebPage url;wsh;⎕IO
    ⍝ Fires up the default browser and goes to "url"
    ⍝ For displaying a file rather then a url add "file://".
    ⍝ However, note that "file://" does not work on some systems.
    ⍝ Examples:
    ⍝ GoToWebPage 'file:///c:/my.html'
    ⍝ GoToWebPage 'file://localhost/c:/my.html'
    ⍝ The to calls are equivalent
     
      ⎕WX←1 ⋄ ⎕IO←0 ⋄ ⎕ML←3
      r←⍬
      :If ~∨/'file://'⍷Lowercase url
          url←'http://'{((⍺≢Lowercase(⍴⍺)↑⍵)/⍺),⍵}url
      :EndIf
      'wsh'⎕WC'OLEClient' 'WScript.Shell'
      {}wsh.Run url
    ∇

      FindPathTo←{
      ⍝ Tries to find ⍵ (typically the name of a needed script) in:
      ⍝ 1. The same namespace ⎕THIS is coming from
      ⍝ 2. The namespaces where the class (if instanciated) comes from
      ⍝ 3. In #
      ⍝ 4. Where it was called from (NOT the same as 1 and 2)
      ⍝ If it fails to find ⍵, an empty string is returned.
          ⎕IO←1 ⋄ ⎕ML←1
          base←⎕THIS.##                    ⍝ were are we coming from?
          0<base.⎕NC ⍵:base                ⍝ Is it in that dir?
          base←(1⊃∊⎕CLASS ⎕THIS).##        ⍝ Where stems the class from?
          0<#.⎕NC ⍵:#                      ⍝ Is it in root?
          0::''                            ⍝ To be on the save side (may be called from inside an instance)
          path←⍎1⊃⎕NSI~⊂⍕⎕THIS             ⍝ From where got we called?
          0<path.⎕NC ⍵:path                ⍝ May be it's there?
          ''                               ⍝ Give up
      }

    First←{⎕ML←3 ⋄ ↑⍵}                     ⍝ "First" ⎕ml independent
    Mix←{⎕ML←0 ⋄ ↑⍵}                       ⍝ "Mix" ⎕ml independent
    Where←{⎕IO←(⎕IO⊃⎕RSI).⎕IO ⋄ ⍵/⍳⍴,⍵}    ⍝ Return indices for Boolean ⍵; depends on ⎕IO in caller's space

:EndNamespace ⍝ APLTreeUtils